import os
import yaml
import shutil
from pdf_writer import PdfWriter
from serializer import doc_to_object, object_to_doc

BASE_PATH = './out'

DOC_PATH = 'doc'
PDF_PATH = 'pdf'

PENDING_PATH = 'pending'
SENT_PATH = 'sent'
PAID_PATH = 'paid'

YAML_EXTENSION = '.yaml'
PDF_EXTENSION = '.pdf'


class DocManager(object):
    def __init__(self, resources):
        self._resources = resources

    def make_pending(self, doc):
        doc_dir_path = _prepare_dir(DOC_PATH, PENDING_PATH)
        pdf_dir_path = _prepare_dir(PDF_PATH, PENDING_PATH)

        _write_doc_to_yaml(doc_dir_path, doc)

        pdf_path = os.path.join(pdf_dir_path, _create_doc_pdf_name(doc.number))
        writer = PdfWriter(self._resources, pdf_path, doc)
        writer.write()

        print("%s was created" % doc.number)

    def move_to_sent(self, doc_number):
        _move(PENDING_PATH, SENT_PATH, doc_number)
        print("%s was moved to sent state" % doc_number)

    def move_to_paid(self, doc_number):
        _move(SENT_PATH, PAID_PATH, doc_number)
        print("%s was moved to paid state" % doc_number)

    def clear_pending(self):
        shutil.rmtree(_prepare_dir(DOC_PATH, PENDING_PATH))
        shutil.rmtree(_prepare_dir(PDF_PATH, PENDING_PATH))
        print("Pending were cleared")

    def get_latest_doc_name(self):
        doc_sent_path = _prepare_dir(DOC_PATH, SENT_PATH)
        doc_paid_path = _prepare_dir(DOC_PATH, PAID_PATH)
        doc_files = [_strip_yaml_from_name(f) for f in os.listdir(doc_sent_path)] + \
                    [_strip_yaml_from_name(f) for f in os.listdir(doc_paid_path)]

        if len(doc_files) > 0:
            doc_files.sort()
            return doc_files[len(doc_files) - 1]
        else:
            return None


def _prepare_dir(type, state):
    dir_path = os.path.join(BASE_PATH, type, state)
    if not os.path.exists(dir_path):
        os.makedirs(dir_path)
    return dir_path


def _write_doc_to_yaml(path, doc):
    with open(os.path.join(path, _create_doc_yaml_name(doc.number)), 'w') as outfile:
        yaml.dump(doc_to_object(doc), outfile, default_flow_style=False)


def _read_doc_from_yaml(path, doc_number):
    with open(os.path.join(path, _create_doc_yaml_name(doc_number)), 'r') as infile:
        obj = yaml.load(infile)
    return object_to_doc(obj)


def _create_doc_yaml_name(doc_number):
    return str(doc_number) + YAML_EXTENSION


def _strip_yaml_from_name(file_name):
    return file_name[:-len(YAML_EXTENSION)]


def _create_doc_pdf_name(doc_number):
    return str(doc_number) + PDF_EXTENSION


def _move(from_state, to_state, doc_number):
    shutil.move(
        os.path.join(_prepare_dir(DOC_PATH, from_state), _create_doc_yaml_name(doc_number)),
        os.path.join(_prepare_dir(DOC_PATH, to_state), _create_doc_yaml_name(doc_number))
    )
    shutil.move(
        os.path.join(_prepare_dir(PDF_PATH, from_state), _create_doc_pdf_name(doc_number)),
        os.path.join(_prepare_dir(PDF_PATH, to_state), _create_doc_pdf_name(doc_number))
    )



from model import *
from datetime import datetime, timedelta


def _last_day_of_month(any_date):
    next_month = any_date.replace(day=28) + timedelta(days=4)  # move to next month - this will never fail
    return next_month - timedelta(days=next_month.day)


def _last_day_of_previous_month(any_date):
    return any_date.replace(day=1) + timedelta(days=-1)  # move to previous month


class DocumentBuilder(object):
    def __init__(self, base, resources, latest_doc_name):
        self.base = Base(base)
        self.resources = resources
        self.latest_doc_name = latest_doc_name

    def build(self, records, date=None):
        if date is None:
            today = datetime.now()
            if today.day < 15:
                date = _last_day_of_previous_month(today)
            else:
                date = _last_day_of_month(today)

        docs = []

        # create temporary documents and sort them base on client order
        temp_docs = []
        for (client_key, client) in self.base.clients.iteritems():
            items = self._find_items(records, client)
            if items is None:
                continue
            items.sort(key=lambda i: i.name)
            discount = self._find_discount(records, client_key)
            temp_docs.append(TempDoc(client, items, discount))
        temp_docs = sorted(temp_docs, key=lambda c: c.client.order)

        # create documents
        for temp_doc in temp_docs:
            # ignore documents with no item
            if len(temp_doc.items) < 1:
                continue

            doc_name = self._check_format_doc_name(date, len(docs) + 1)
            docs.append(Document(
                number=doc_name,

                issue_date=self._format_date(date),
                delivery_date=self._format_date(date),
                due_date=self._format_date(date + timedelta(days=14)),

                contractor=self.base.contractor,
                currency=self.base.default_curremcy,
                bank_account=self.base.bank_accounts[self.base.default_bank_account_name],

                client=temp_doc.client,
                discount=temp_doc.discount,
                items=temp_doc.items
            ))

        return docs

    def _format_doc_name(self, date, num):
        numbering = self.base.numbering
        if numbering == 'YYYYMMccc':
            return "{:%Y%m}{:03d}".format(date, num)
        elif numbering == 'YYYYccc':
            return "{:%Y}{:03d}".format(date, num)
        else:
            return None

    def _check_format_doc_name(self, date, num):
        n = self._format_doc_name(date, num)
        if self.latest_doc_name is None or n[:-3] != self.latest_doc_name[:-3]:
            return n
        else:
            latest = int(self.latest_doc_name[-3:])
            return self._format_doc_name(date, num + latest)

    def _format_date(self, date):
        return "{:%d.%m.%Y}".format(date)

    def _find_discount(self, records, client_name):
        discount = 0
        used_records = []

        for record in records:
            if isinstance(record, DiscountRecord):
                if record.client_name == client_name:
                    discount += record.price
                    used_records.append(record)

        for record in used_records:
            records.remove(record)
        return discount

    def _find_items(self, records, client):
        items = []
        used_records = []

        for record in records:
            if isinstance(record, PriceRecord) or isinstance(record, HoursRecord):
                for (_, project) in client.projects.iteritems():
                    if record.project_name in project.aliases:
                        items.append(self._map_record_to_item(project, record))
                        used_records.append(record)
                        break

        for record in used_records:
            records.remove(record)
        return self._combine_same_items(items)

    def _map_record_to_item(self, project, record):
        if isinstance(record, HoursRecord):
            return Item(
                name=self.resources.get_string('item_name_' + project.type, project.name),
                amount=record.hours,
                unit=Unit.HOUR,
                price=project.price
            )
        elif isinstance(record, PriceRecord):
            return Item(
                name=self.resources.get_string('item_name_ks', project.name),
                amount=1,
                unit=Unit.KS,
                price=record.price
            )

    def _combine_same_items(self, items_list):
        items_map = {}

        for item in items_list:
            if item.name in items_map:
                old = items_map[item.name]
                if item.unit == Unit.HOUR:
                    items_map[item.name] = Item(
                        name=item.name,
                        price=item.price,
                        amount=item.amount + old.amount,
                        unit=item.unit
                    )
                else:
                    items_map[item.name] = Item(
                        name=item.name,
                        price=item.price + old.price,
                        amount=item.amount,
                        unit=item.unit
                    )
            else:
                items_map[item.name] = item

        return items_map.values()


class TempDoc(object):
    def __init__(self, client, items, discount):
        self.client = client
        self.items = items
        self.discount = discount



from model import *


def doc_to_object(doc):
    items = []
    for item in doc.items:
        items.append({
            'name': item.name,
            'amount': item.amount,
            'unit': item.unit,
            'price': item.price
        })

    return {
        'number': doc.number,
        'due_date': doc.due_date,
        'issue_date': doc.issue_date,
        'delivery_date': doc.delivery_date,

        'contractor': {
            'name': doc.contractor.name,
            'address': doc.contractor.address,
            'ico': doc.contractor.ico,
            'dic': doc.contractor.dic,
            'icdph': doc.contractor.icdph,
            'phone': doc.contractor.phone,
            'email': doc.contractor.email,
            'signature': doc.contractor.signature
        },
        'bank_account': {
            'iban': doc.bank_account.iban,
            'swift': doc.bank_account.swift
        },
        'currency': doc.currency,

        'client': {
            'name': doc.client.name,
            'address': doc.client.address,
            'ico': doc.client.ico,
            'dic': doc.client.dic,
            'icdph': doc.client.icdph,
            'price': doc.client.price
        },
        'discount': doc.discount,
        'items': items
    }


def object_to_doc(obj):
    items = []
    for i in obj['items']:
        items.append(Item(i['name'], i['amount'], i['unit'], i['price']))

    return Document(
        number=obj['number'],

        issue_date=obj['issue_date'],
        delivery_date=obj['delivery_date'],
        due_date=obj['due_date'],

        contractor=Contractor(obj['contractor']),
        currency=obj['currency'],
        bank_account=BankAccount(obj['bank_account']),

        client=Client(obj['client']),
        discount=obj['discount'],
        items=items
    )



import argparse

from util import load_yaml
from resource import ResourceManager
from builder import DocumentBuilder
from doc_manager import DocManager
import input_parser
import os


def run(resources):
    doc_manager = DocManager(resources)
    latest_doc_name = doc_manager.get_latest_doc_name()

    # retrieve input
    # TODO : tempo timesheets API : https://tempoplugin.jira.com/wiki/spaces/JTS/pages/120389685/Timesheets+REST+API
    records = input_parser.parse(os.path.abspath('IN'))

    # prepare documents
    builder = DocumentBuilder(load_yaml('base.yaml'), resources, latest_doc_name)
    docs = builder.build(records)

    # create documents
    doc_manager.clear_pending()
    for doc in docs:
        doc_manager.make_pending(doc)


def sent(resources, doc_number):
    doc_manager = DocManager(resources)
    doc_manager.move_to_sent(doc_number)


def paid(resources, doc_number):
    doc_manager = DocManager(resources)
    doc_manager.move_to_paid(doc_number)


def main(debug=False):
    parser = argparse.ArgumentParser()
    parser.add_argument('-r', '--run', action='store_true',
                        help="Create new documents")
    parser.add_argument('-s', '--sent', nargs=1, metavar="DOC_NUMBER",
                        help="Move the document with given number to the sent state")
    parser.add_argument('-p', '--paid', nargs=1, metavar="DOC_NUMBER",
                        help="Move the document with given number to the paid state")
    args = parser.parse_args()

    if args.run:
        run(ResourceManager(debug))
    elif args.sent is not None:
        sent(ResourceManager(debug), args.sent[0])
    elif args.paid is not None:
        paid(ResourceManager(debug), args.paid[0])


if __name__ == "__main__":
    main(debug=True)



#####################################################################################
# Base
#####################################################################################
class CurrencyImpl(object):
    def __init__(self, name, sign):
        self.name = name
        self.sign = sign


class Currency(object):
    EUR = CurrencyImpl('EUR', u'\u20ac')
    NONE = CurrencyImpl('', u'')

    all = [EUR]

    @staticmethod
    def for_name(name):
        for currency in Currency.all:
            if currency.name == name:
                return currency
        return Currency.NONE


class ProjectType(object):
    MOBILE = 'mobile'
    DESKTOP = 'desktop'

    all = [MOBILE, DESKTOP]

    @staticmethod
    def validate(name):
        if name in ProjectType.all:
            return name
        else:
            return None


class Base(object):
    def __init__(self, data):
        self.contractor = Contractor(data)
        self.numbering = data['numbering']
        self.default_curremcy = Currency.for_name(data['default_curremcy'])
        self.default_bank_account_name = data['default_bank_account']
        self.bank_accounts = {}
        for (acc_name, acc_data) in data['bank_accounts'].iteritems():
            self.bank_accounts[acc_name] = BankAccount(acc_data)

        self.clients = {}
        for (client_name, client_data) in data['clients'].iteritems():
            self.clients[client_name] = Client(client_data)


class Contractor(object):
    def __init__(self, data):
        self.name = data['name']
        self.address = data['address']
        self.ico = data['ico']
        self.dic = data['dic']
        self.icdph = data['icdph']
        self.phone = data['phone']
        self.email = data['email']
        self.signature = data['signature']


class Client(object):
    def __init__(self, data):
        self.name = data['name']
        self.order = data.get('order', 0)
        self.address = data['address']
        self.ico = data['ico']
        self.dic = data['dic']
        self.icdph = data['icdph']
        self.price = data['price']
        self.projects = {}
        for (project_name, project_data) in data.get('projects', {}).iteritems():
            self.projects[project_name] = Project(project_name, project_data, default_price=self.price)

    def is_foreign(self):
        return self.address[len(self.address) - 1] != 'Slovensko'


class Project(object):
    def __init__(self, name, data, default_price):
        self.name = name
        self.type = ProjectType.validate(data['type'])
        self.aliases = data.get('aliases', []) + [name]
        self.price = data.get('price', default_price)


class BankAccount(object):
    def __init__(self, data):
        self.iban = data['iban']
        self.swift = data['swift']


#####################################################################################
# Document
#####################################################################################
class Document(object):
    def __init__(self, number, due_date, issue_date, delivery_date, contractor, bank_account, currency, client, items,
                 discount=0):
        self.number = number
        self.due_date = due_date
        self.issue_date = issue_date
        self.delivery_date = delivery_date

        self.contractor = contractor
        self.bank_account = bank_account
        self.currency = currency

        self.client = client
        self.discount = discount
        self.items = items

    def total(self):
        return self._item_total() - self.discount

    def _item_total(self):
        sum = 0
        for item in self.items:
            sum += item.total()
        return sum

    def discount_percentage(self):
        return self.discount / self._item_total() * 100


class Item(object):
    def __init__(self, name, amount, unit, price):
        self.name = name
        self.amount = amount
        self.unit = unit
        self.price = price

    def total(self):
        return self.amount * self.price


class Unit(object):
    KS = 'ks'
    HOUR = 'hod'


#####################################################################################
# Input Record
#####################################################################################


class Record(object):
    pass


class DiscountRecord(Record):
    def __init__(self, client_name, price):
        self.client_name = client_name
        self.price = price


class HoursRecord(Record):
    def __init__(self, project_name, hours):
        self.project_name = project_name
        self.hours = hours


class PriceRecord(Record):
    def __init__(self, project_name, price):
        self.project_name = project_name
        self.price = price



import csv
from model import *

HOUR_CHARACTER = 'h'


def parse(file_path):
    records = []

    with open(file_path, 'rb') as csvfile:
        rows = csv.reader(csvfile, delimiter=' ', quotechar='|')
        for row in rows:
            parts_count = len(row)
            if parts_count < 2:
                continue

            number = float(row[0])

            if number == 0:
                continue
            elif number < 0:
                records.append(DiscountRecord(client_name=create_name(row, 1), price=-number))
            else:  # number > 0
                second_word = str(row[1])
                if second_word == HOUR_CHARACTER:
                    records.append(HoursRecord(project_name=create_name(row, 2), hours=number))
                else:
                    records.append(PriceRecord(project_name=create_name(row, 1), price=number))

    return records


def create_name(words, start_index):
    return ' '.join(words[start_index:])
